"""
@file
@brief Functions implemented einsum computation for two
matrices having the same dimensions.
"""
import numpy


def numpy_diagonal(m, axis, axes):
    """
    Extracts diagonal coefficients from an array.

    :param m: input array
    :param axis: kept axis among the diagonal ones
    :param axes: diagonal axes (axis must be one of them)
    :return: output

    .. runpython::
        :showcode:

        import numpy
        from mlprodict.testing.einsum import numpy_diagonal

        mat = numpy.arange(8).reshape((2, 2, 2))
        print(mat)
        diag = numpy_diagonal(mat, 1, [1, 2])
        print(diag)
    """
    if axis not in axes:
        raise RuntimeError(
            "axis %r must be in axes %r." % (axis, axes))
    shape = []
    new_shape = []
    for i, s in enumerate(m.shape):
        if i in axes:
            if i == axis:
                shape.append(s)
                new_shape.append(s)
            else:
                shape.append(1)
        else:
            shape.append(s)
            new_shape.append(s)

    # Extracts coefficients.
    output = numpy.empty(tuple(shape), dtype=m.dtype)
    index_in = [slice(s) for s in m.shape]
    index_out = [slice(s) for s in m.shape]
    for i in range(0, shape[axis]):
        for a in axes:
            index_in[a] = i
            index_out[a] = i if a == axis else 0
        output[tuple(index_out)] = m[tuple(index_in)]

    # Removes axis.
    return output.reshape(tuple(new_shape))


def _numpy_extended_dot_equation(m1_dim, m2_dim, axes, left, right):
    """
    Returns the equation equivalent to an extended version
    of an aligned matrix multiplication
    (see @see fn numpy_extended_dot).

    :param m1: number of dimensions of the first matrix
    :param m2: number of dimensions of the second matrix
    :param axes: summation axes
    :param axes: summation axes
    :param left: left axes
    :param right: right axes
    :return: equation

    .. runpython::
        :showcode:

        import numpy
        from mlprodict.testing.einsum.einsum_impl_ext import (
            numpy_extended_dot_python, _numpy_extended_dot_equation)

        a = numpy.arange(6).reshape((3, 2, 1))
        b = numpy.arange(12).reshape((3, 1, 4))

        print(numpy_extended_dot_python(
            a, b, axes=(0, ), left=(1,), right=(2,)))

        # Equivalent einsum equation
        print('equation', _numpy_extended_dot_equation(
            len(a.shape), len(a.shape), axes=(0, ), left=(1,), right=(2,)))

        # Same einsum computation written in a different way.
        print(numpy.einsum('kix,kxj->xij', a, b))
    """
    if m1_dim != m2_dim:
        raise RuntimeError(
            "Matrices m1 and m2 must have the same number of dimensions, "
            "m1=%r, m2=%r." % (m1_dim, m2_dim))
    total = set(axes) | set(left) | set(right)
    if len(total) > m1_dim:
        raise ValueError(
            "Whole set of involved axes should be inferior to the number "
            "of dimensions: %r = {%r} | {%r} | {%r} has more than %d elements"
            "." % (total, axes, left, right, m1_dim))

    def _check_(axs, n):
        for a in axs:
            if a < 0 or a >= n:
                raise ValueError(
                    "One axis %d (in %r) is negative or above the maximum "
                    "dimension %d." % (a, axs, n))
    _check_(axes, m1_dim)
    _check_(left, m1_dim)
    _check_(right, m1_dim)

    l1 = [chr(i + 97) for i in range(m1_dim)]
    l2 = [chr(i + 97) for i in range(m1_dim)]
    l3 = [chr(i + 97) for i in range(m1_dim)]
    for a in left:
        l1[a] = l1[a].upper()
        l3[a] = l3[a].upper()
    for a in right:
        l2[a] = l2[a].upper()
        l3[a] = l3[a].upper()
    for a in axes:
        l1[a] = l1[a].lower()
        l2[a] = l2[a].lower()
        if a not in right:
            l3[a] = None
        else:
            l3[a] = l3[a].lower()
    eq = "%s,%s->%s" % ("".join(l1), "".join(l2),
                        "".join(s for s in l3 if s))
    return eq


def _common_check_numpy_extended_dot(m1, m2, axes, left, right):
    """
    Common verifications for all implementations of
    @see fn numpy_extended_dot.
    """
    if m1.dtype != m2.dtype:
        raise TypeError(
            "Both matrices should share the same dtype %r != %r."
            "" % (m1.dtype, m2.dtype))
    m1_dim = len(m1.shape)
    m2_dim = len(m2.shape)
    if m1_dim != m2_dim:
        raise RuntimeError(  # pragma: no cover
            "Matrices m1 and m2 must have the same number of dimensions, "
            "m1=%r, m2=%r." % (m1_dim, m2_dim))
    total = set(axes) | set(left) | set(right)
    if len(total) > m1_dim:
        raise ValueError(
            "Whole set of involved axes should be inferior to the number "
            "of dimensions: %r = {%r} | {%r} | {%r} has more than %d elements"
            "." % (total, axes, left, right, m1_dim))


def numpy_extended_dot(m1, m2, axes, left, right, verbose=False):
    """
    Extended version of a matrix multiplication (:epkg:`numpy:dot`)
    with two matrices *m1*, *m2* of the same dimensions.
    Loops over *left* axes for *m1* and *right* axes for *m2*,
    summation is done over *axes*.
    Other axes must be empty.
    This multiplication combines matrix multiplication (dot)
    and broadcasted multiplication term by term.

    :param m1: first matrix
    :param m2: second matrix
    :param axes: summation axes
    :param left: left axes
    :param right: right axes
    :param verbose: display intermediate information
    :return: output

    The dot product is equivalent to:

    .. runpython::
        :showcode:

        import numpy
        from mlprodict.testing.einsum import numpy_extended_dot

        m1 = numpy.arange(4).reshape((2, 2))
        m2 = m1 + 10
        print("dot product")
        print(m1 @ m2)

        dm1 = m1.reshape((2, 2, 1))
        dm2 = m2.reshape((1, 2, 2))
        dot = numpy_extended_dot(dm1, dm2, axes=[1], left=[0], right=[2],
                                 verbose=True)
        print("extended dot product")
        print(dot)

    Empty axes should be squeezed to get identical results.
    Dot product when the second matrix is transposed.

    .. runpython::
        :showcode:

        import numpy
        from mlprodict.testing.einsum import numpy_extended_dot

        m1 = numpy.arange(4).reshape((2, 2))
        m2 = m1 + 10
        print("dot product")
        print(m1 @ m2.T)

        dm1 = m1.reshape((2, 1, 2))
        dm2 = m2.reshape((1, 2, 2))
        dot = numpy_extended_dot(dm1, dm2, axes=[2], left=[0], right=[1],
                                 verbose=True)
        print("extended dot product")
        print(dot)

    An example when right axes include the summation axis.

    .. runpython::
        :showcode:

        import numpy
        from mlprodict.testing.einsum import numpy_extended_dot

        m1 = numpy.arange(4).reshape((2, 2))
        m2 = m1 + 10
        dm1 = m1.reshape((2, 2, 1))
        dm2 = m2.reshape((1, 2, 2))
        dot = numpy_extended_dot(dm1, dm2, axes=[2], left=[0], right=[1, 2],
                                 verbose=True)
        print(dot)

    Example in higher dimension:

    .. runpython::
        :showcode:

        import numpy
        from mlprodict.testing.einsum import numpy_extended_dot

        m1 = numpy.arange(8).reshape((2, 2, 2))
        m2 = m1 + 10

        dot = numpy_extended_dot(m1, m2, [1], [0], [2], verbose=True)
        print(dot)

    The current implementation still uses :epkg:`numpy:einsum`
    but this should be replaced.
    """
    _common_check_numpy_extended_dot(m1, m2, axes, left, right)
    eq = _numpy_extended_dot_equation(
        len(m1.shape), len(m2.shape), axes, left, right)
    if verbose:
        print("  [numpy_extended_dot] %s: %r @ %r" % (eq, m1.shape, m2.shape))
    output = numpy.einsum(eq, m1, m2)
    new_shape = list(output.shape)
    for a in axes:
        if a not in right:
            new_shape.insert(a, 1)
    if verbose:
        print("  [numpy_extended_dot] %r reshaped into %r " % (
            output.shape, new_shape))
    return output.reshape(tuple(new_shape))


def numpy_extended_dot_ouput_shape(m1, m2, axes, left, right):
    """
    Computes the output shape of results produced by function
    :func:`numpy_extended_dot
    <mlprodict.testing.einsum_impl_ext.numpy_extended_dot>` or
    :func:`numpy_extended_dot_python
    <mlprodict.testing.einsum_impl_ext.numpy_extended_dot_python>`.
    """
    _common_check_numpy_extended_dot(m1, m2, axes, left, right)
    m1_dim = len(m1.shape)

    new_shape = numpy.full(m1_dim, 1, dtype=numpy.int64)
    for i in left:
        new_shape[i] = m1.shape[i]
    for i in right:
        if (i in left and m1.shape[i] != m2.shape[i] and
                m1.shape[i] != 1 and m2.shape[i] != 1):
            raise RuntimeError(  # pragma: no cover
                "Matrices should have the same dimension for dimension %d, "
                "shapes=%r @ %r." % (i, m1.shape, m2.shape))
        new_shape[i] = m2.shape[i]
    return new_shape


def _numpy_extended_dot_python_l1l2l3(m1_dim, axes, left, right):
    l1 = [chr(i + 97) for i in range(m1_dim)]
    l2 = [chr(i + 97) for i in range(m1_dim)]
    l3 = [chr(i + 97) for i in range(m1_dim)]
    for a in left:
        l1[a] = l1[a].upper()
        l3[a] = l3[a].upper()
    for a in right:
        l2[a] = l2[a].upper()
        l3[a] = l3[a].upper()
    for a in axes:
        l1[a] = l1[a].lower()
        l2[a] = l2[a].lower()
        if a not in right:
            l3[a] = "-"
        else:
            l3[a] = l3[a].lower()
    return l1, l2, l3


def _numpy_extended_dot_python_intermediate(m1_shape, m2_shape, l1, l2, l3):
    names = list(sorted(set(l1 + l2)))
    kind = numpy.zeros(len(names), dtype=numpy.int64)
    cols = {}

    for i, n in enumerate(names):
        if n in l1:
            kind[i] += 1
            cols[n] = l1.index(n)
        if n in l2:
            kind[i] += 2
            cols[n] = l2.index(n)
        if n in l3:
            kind[i] += 4

    pos = numpy.zeros(len(names), dtype=numpy.int64)
    for j in range(0, pos.shape[0]):
        pos[j] = cols[names[j]]
    common = [(kind[i] & 3) == 3 for i in range(len(kind))]
    broadcast = [common[i] and m1_shape[pos[i]] != m2_shape[pos[i]]
                 for i in range(len(common))]

    return names, kind, cols, common, broadcast, pos


def _numpy_extended_dot_python_update_broadcast(
        m1, m2, axes, left, right, l1, l2, l3, names, broadcast, cols,
        kind, common, verbose=False):

    def dispb(c):
        return "".join("o" if b else "." for b in c)

    if verbose:
        print("[GENERICDOT] before broadcast %s,%s->%s      or %s" % (
            "".join(l1), "".join(l2), "".join(l3),
            _numpy_extended_dot_equation(
                len(m1.shape), len(m1.shape), axes, left, right)))
        print("[GENERICDOT] names=%s kind=%r common=%s broadcast=%s" % (
            "".join(names), kind.tolist(),
            dispb(common), dispb(broadcast)))

    for i in range(len(broadcast)):  # pylint: disable=C0200
        if broadcast[i] and not (kind[i] & 3) == 3:
            raise RuntimeError(  # pragma: no cover
                "Broadcast should only happen on common axes, "
                "axes=%r left=%r right=%r shape1=%r shape2=%r."
                "" % (axes, left, right, m1.shape, m2.shape))
        if not broadcast[i]:
            continue
        # We split letters.
        p = cols[names[i]]
        dim = (m1.shape[p], m2.shape[p])
        let = [l1[p], l2[p], l3[p]]
        inp = 1 if dim[0] == 1 else 0
        if verbose:
            print("[GENERICDOT] name=%s dim=%r let=%r inp=%r p=%r" % (
                names[i], dim, let, inp, p))
            print("    B0 l1=%r, l2=%r l3=%r" % (l1, l2, l3))
        if (kind[i] & 4) > 0:
            # Summation axis is part of the output.
            if let[inp].lower() == let[inp]:
                let[inp] = let[inp].upper()
            else:
                let[inp] = let[inp].lower()
            l3[p] = let[inp]
            if inp == 1:
                l2[p] = let[inp]
            else:
                l1[p] = let[inp]
            if verbose:
                print("    B1 l1=%r, l2=%r l3=%r" % (l1, l2, l3))
        else:
            # Summation axis is not part of the output.
            if let[inp].lower() == let[inp]:
                let[inp] = let[inp].upper()
            else:
                let[inp] = let[inp].lower()
            if inp == 1:
                l2[p] = let[inp]
            else:
                l1[p] = let[inp]
            if verbose:
                print("    B2 l1=%r, l2=%r l3=%r" % (l1, l2, l3))

    return l1, l2, l3


def numpy_extended_dot_python(m1, m2, axes, left, right, verbose=False):
    """
    Implementation of @see fn numpy_extended_dot in pure python.
    This implementation is not efficient but shows how to
    implement this operation without :epkg:`numpy:einsum`.

    .. runpython::
        :showcode:

        import numpy
        from mlprodict.testing.einsum import numpy_extended_dot_python
        from mlprodict.testing.einsum.einsum_impl_ext import (
            _numpy_extended_dot_equation)

        a = numpy.arange(6).reshape((3, 2, 1))
        b = numpy.arange(12).reshape((3, 1, 4))

        print(numpy_extended_dot_python(
            a, b, axes=(0, ), left=(1,), right=(2,)))

        # Equivalent einsum equation
        print('equation', _numpy_extended_dot_equation(
            len(a.shape), len(a.shape), axes=(0, ), left=(1,), right=(2,)))

        # Same einsum computation written in a different way.
        print(numpy.einsum('kix,kxj->xij', a, b))
    """
    def dispb(c):
        return "".join("o" if b else "." for b in c)

    new_shape = numpy_extended_dot_ouput_shape(m1, m2, axes, left, right)
    m1_dim = len(m1.shape)

    # output result
    res = numpy.full(tuple(new_shape), 0, dtype=m1.dtype)

    # indices
    l1, l2, l3 = _numpy_extended_dot_python_l1l2l3(m1_dim, axes, left, right)
    names, kind, cols, common, broadcast, pos = (
        _numpy_extended_dot_python_intermediate(
            m1.shape, m2.shape, l1, l2, l3))

    if any(broadcast):
        l1, l2, l3 = _numpy_extended_dot_python_update_broadcast(
            m1, m2, axes, left, right, l1, l2, l3, names, broadcast, cols,
            kind, common, verbose=verbose)

        names, kind, cols, common, broadcast, pos = (
            _numpy_extended_dot_python_intermediate(
                m1.shape, m2.shape, l1, l2, l3))

    indices = numpy.array([0 for n in names], dtype=numpy.int64)
    pl1 = numpy.array([names.index(c) for c in l1], dtype=numpy.int64)
    pl2 = numpy.array([names.index(c) for c in l2], dtype=numpy.int64)
    limits = numpy.array(
        [m1.shape[pos[n]] if (kind[n] & 1) == 1 else m2.shape[pos[n]]
         for n in range(len(names))], dtype=numpy.int64)
    plo = numpy.array(
        [-1 if c not in names else names.index(c) for c in l3],
        dtype=numpy.int64)

    if verbose:
        print("[GENERICDOT] %s,%s->%s      or %s" % (
            "".join(l1), "".join(l2), "".join(l3),
            _numpy_extended_dot_equation(
                len(m1.shape), len(m1.shape), axes, left, right)))
        print("[GENERICDOT] shape1=%r shape2=%r shape=%r" % (
            m1.shape, m2.shape, res.shape))
        print("[GENERICDOT] axes=%r left=%r right=%r" % (axes, left, right))
        print("[GENERICDOT] pl1=%r pl2=%r plo=%r" % (pl1, pl2, plo))
        print("[GENERICDOT] names=%s kind=%r common=%s broadcast=%s" % (
            "".join(names), kind.tolist(),
            dispb(common), dispb(broadcast)))
        print("[GENERICDOT] pos=%r" % pos.tolist())
        print("[GENERICDOT] cols=%r" % cols)
        print("[GENERICDOT] limits=%r" % limits)

    while indices[0] < limits[0]:

        # The function spends most of its time is these three lines.
        t1 = tuple(indices[n] for n in pl1)
        t2 = tuple(indices[n] for n in pl2)
        to = tuple(0 if n == -1 else indices[n] for n in plo)

        c = m1[t1] * m2[t2]

        if verbose:
            print(" %r x %r -> %r v=%r I=%r" % (t1, t2, to, c, indices))

        res[to] += c

        last = len(indices) - 1
        indices[last] += 1
        for i in range(last, 0, -1):
            if indices[i] < limits[i]:
                break
            indices[i] = 0
            if i > 0:
                indices[i - 1] += 1

    return res


def numpy_extended_dot_matrix(m1, m2, axes, left, right, verbose=False):
    """
    Implementation of @see fn numpy_extended_dot using dot product,
    multiplication, transpose and reduction
    but not a custom python implementation like
    @see fn numpy_extended_dot_python.

    .. runpython::
        :showcode:

        import numpy
        from mlprodict.testing.einsum import numpy_extended_dot_matrix
        from mlprodict.testing.einsum.einsum_impl_ext import (
            _numpy_extended_dot_equation)

        a = numpy.arange(6).reshape((3, 2, 1))
        b = numpy.arange(12).reshape((3, 1, 4))

        print(numpy_extended_dot_matrix(
            a, b, axes=(0, ), left=(1,), right=(2,)))

        # Equivalent einsum equation
        print('equation', _numpy_extended_dot_equation(
            len(a.shape), len(a.shape), axes=(0, ), left=(1,), right=(2,)))

        # Same einsum computation written in a different way.
        print(numpy.einsum('kix,kxj->xij', a, b))
    """
    _common_check_numpy_extended_dot(m1, m2, axes, left, right)

    if verbose:
        print("[GENERICDOT] shape1=%r shape2=%r axes=%r "
              "left=%r right=%r -- %s" % (
                  m1.shape, m2.shape, axes, left, right,
                  _numpy_extended_dot_equation(
                      len(m1.shape), len(m1.shape), axes, left, right)))

    if len(axes) == 0 and len(set(left) & set(right)) == 0:
        # Simple multiplication
        res = m1 * m2
        if verbose:
            print("[GENERICDOT] Mul %r @ %r -> %r" % (
                m1.shape, m2.shape, res.shape))
        return res

    if (len(set(axes) & set(left)) == 0 and
            len(set(axes) & set(right)) == 0):

        # No intersection between axes and right: matrix multiplication
        # ReduceSum
        right_no_left = set(right) - (set(right) & (set(left) | set(axes)))
        if right_no_left:
            red1 = m1.sum(axis=tuple(sorted(right_no_left)), keepdims=True)
            if verbose:
                print("[GENERICDOT] reducesumL=%r, %r -> %r" % (
                    right_no_left, m1.shape, red1.shape))
        else:
            red1 = m1

        left_no_right = set(left) - (set(left) & (set(right) | set(axes)))
        if left_no_right:
            red2 = m2.sum(axis=tuple(sorted(left_no_right)), keepdims=True)
            if verbose:
                print("[GENERICDOT] reducesumR=%r, %r -> %r" % (
                    left_no_right, m2.shape, red2.shape))
        else:
            red2 = m2

        # Transpose
        common_axes = sorted(set(left) & set(right))
        i_axes = [(-1 if i in common_axes
                   else (1 if i in axes else 0), i)
                  for i in range(len(m1.shape))]
        i_axes.sort()
        perm = [_[1] for _ in i_axes]
        trm1 = numpy.transpose(red1, axes=perm)
        trm2 = numpy.transpose(red2, axes=perm)
        if verbose:
            print("[GENERICDOT] transposeL=%r, %r -> %r" % (
                perm, red1.shape, trm1.shape))
            print("[GENERICDOT] transposeR=%r, %r -> %r" % (
                perm, red2.shape, trm2.shape))
        final_shape = numpy_extended_dot_ouput_shape(
            m1, m2, axes, left, right)
        perm_left = [i for i in range(len(perm)) if perm[i] in left]
        perm_right = [i for i in range(len(perm)) if perm[i] in right]
        perm_common_axes = [i for i in range(len(perm))
                            if perm[i] in common_axes]

        if verbose:
            print("[GENERICDOT] MatMul %r @ %r -> %r  --  %s" % (
                m1.shape, m2.shape, final_shape,
                _numpy_extended_dot_equation(
                    len(m1.shape), len(m1.shape), axes, left, right)))
            print("[GENERICDOT] axes=%r left=%r right=%r" %
                  (axes, left, right))
            print("[GENERICDOT] perm=%r perm_left=%r "
                  "perm_right=%r perm_common_axes=%r" % (
                      perm, perm_left, perm_right, perm_common_axes))

        # Reshape
        dim0 = int(numpy.prod([trm1.shape[i] for i in perm_common_axes]))
        dim0b = int(numpy.prod([trm2.shape[i] for i in perm_common_axes]))
        if len(axes) > 0:
            all_axes = list(range(0, len(m1.shape)))
            new_axes = all_axes[-len(axes):]
        else:
            new_axes = []
        dim1 = int(numpy.prod([trm1.shape[i] for i in new_axes]))
        dim2 = int(numpy.prod([trm2.shape[i] for i in new_axes]))
        if dim1 != dim2:
            raise RuntimeError(  # pragma: no cover
                "Summation axis do not have the same length %d != %d, "
                "trshape1=%r trshape2=%r "
                "p_axes=%r p_left=%r p_right=%r p_common=%r"
                "." % (dim1, dim2, trm1.shape, trm2.shape,
                       new_axes, perm_left, perm_right, perm_common_axes))
        else:
            shm1 = trm1.reshape((dim0, -1, dim1))
            shm2 = trm2.reshape((dim0b, -1, dim2))

            if verbose:
                print("[GENERICDOT] Reshape %r @ %r -> %r @ %r" % (
                    (dim0, -1, dim1), (dim0, -1, dim2),
                    shm1.shape, shm2.shape))
                print("[GENERICDOT] matmul")

            # Multiplication (this should be done in a different way.
            res = shm1 @ numpy.transpose(shm2, axes=(0, 2, 1))

        if verbose:
            print("[GENERICDOT] Shape after multiplication %s" % (res.shape, ))

        # Transpose again
        not_in_both = []
        for i in range(0, len(m1.shape)):
            if i not in left and i not in right:
                not_in_both.append(i)
        ordered_axes = (common_axes +
                        list(i for i in left if i not in right) +
                        list(i for i in right if i not in left) +
                        not_in_both)

        perm_not_in_both = [i for i in range(len(perm))
                            if perm[i] in not_in_both]
        current_shape = ([max(trm1.shape[i], trm2.shape[i])
                          for i in sorted(perm_common_axes)] +
                         [trm1.shape[i] for i in sorted(perm_left)
                          if i not in perm_common_axes] +
                         [trm2.shape[i] for i in sorted(perm_right)
                          if i not in perm_common_axes] +
                         [1 for i in perm_not_in_both])

        if verbose:
            print("[GENERICDOT] current_shape=%r final_shape=%r "
                  "last_shape=%r" % (current_shape, final_shape, res.shape))

        if len(current_shape) != len(final_shape):
            raise RuntimeError(
                "Shapes mismatch %r > %r, "
                "shape1=%r shape2=%r axes=%r left=%r right=%r." % (
                    current_shape, final_shape,
                    m1.shape, m2.shape, axes, left, right))

        res = res.reshape(current_shape)

        perm = [(a, i) for i, a in enumerate(ordered_axes)]
        perm.sort()
        perm = [p[1] for p in perm]

        if verbose:
            print("[GENERICDOT] ordered_axes=%r perm=%r" % (
                ordered_axes, perm))

        return numpy.transpose(res, axes=perm)

    else:
        # Multiplication and Matrix multiplication at the same time.
        l_axes = set(left) & set(axes)
        r_axes = set(right) & set(axes)
        if r_axes and not l_axes:
            new_axes = list(a for a in axes if a not in right)
            new_left = list(sorted(set(left) | r_axes))
            if verbose:
                eq1 = _numpy_extended_dot_equation(
                    len(m1.shape), len(m1.shape), axes, left, right)
                eq2 = _numpy_extended_dot_equation(
                    len(m1.shape), len(m1.shape), new_axes, new_left, right)
                print("[GENERICDOT] replace left %r by %r axes %r by %r, "
                      "eq %r by %r" % (
                          left, new_left, axes, new_axes, eq1, eq2))
            return numpy_extended_dot_matrix(m1, m2, new_axes, new_left, right,
                                             verbose=verbose)
        raise RuntimeError(  # pragma: no cover
            "shape1=%r shape2=%r axes=%r left=%r right=%r eq=%s." % (
                m1.shape, m2.shape, axes, left, right,
                _numpy_extended_dot_equation(
                    len(m1.shape), len(m1.shape), axes, left, right)))



"""
@file
@brief Inspired from :epkg:`sklearn-onnx`, handles two backends.
"""
from .utils_backend_onnxruntime import compare_runtime as compare_runtime_ort
from .utils_backend_python import compare_runtime as compare_runtime_pyrt


def compare_backend(backend, test, decimal=5, options=None, verbose=False,
                    context=None, comparable_outputs=None,
                    intermediate_steps=False, classes=None,
                    disable_optimisation=False):
    """
    The function compares the expected output (computed with
    the model before being converted to ONNX) and the ONNX output.

    :param backend: backend to use to run the comparison
    :param test: dictionary with the following keys:
        - *onnx*: onnx model (filename or object)
        - *expected*: expected output (filename pkl or object)
        - *data*: input data (filename pkl or object)
    :param decimal: precision of the comparison
    :param options: comparison options
    :param context: specifies custom operators
    :param comparable_outputs: compare only these outputs
    :param verbose: in case of error, the function may print
        more information on the standard output
    :param intermediate_steps: displays intermediate steps
        in case of an error
    :param classes: classes names (if option 'nocl' is used)
    :param disable_optimisation: disable optimisation onnxruntime
        could do

    The function does not return anything but raises an error
    if the comparison failed.
    :return: tuple (output, lambda function to call onnx predictions)
    """
    if backend == "onnxruntime":
        return compare_runtime_ort(
            test, decimal, options=options, verbose=verbose,
            comparable_outputs=comparable_outputs,
            intermediate_steps=False, classes=classes,
            disable_optimisation=disable_optimisation)
    if backend == "python":
        return compare_runtime_pyrt(
            test, decimal, options=options, verbose=verbose,
            comparable_outputs=comparable_outputs,
            intermediate_steps=intermediate_steps, classes=classes,
            disable_optimisation=disable_optimisation)
    raise ValueError(  # pragma: no cover
        "Does not support backend '{0}'.".format(backend))



"""
@file
@brief Functions to help guessing the final graph structure.
"""
import numpy
try:
    from onnxconverter_common.data_types import Float16TensorType
except ImportError:  # pragma: no cover
    Float16TensorType = None
from skl2onnx.common.data_types import (
    DataType,
    FloatTensorType, SequenceType, DictionaryType,
    Int64Type, Int64TensorType, BooleanTensorType,
    Int32TensorType, DoubleTensorType, FloatType,
    StringTensorType)
from skl2onnx.common.data_types import (
    _guess_type_proto, _guess_type_proto_str)
from skl2onnx.algebra.type_helper import _guess_type as skl2onnx__guess_type
from skl2onnx.proto import TensorProto


def _guess_type(var):
    if isinstance(var, dict) and 'value' in var:
        return skl2onnx__guess_type(var['value'])  # pragma: no cover
    return skl2onnx__guess_type(var)


def get_defined_inputs(input_names, variables=None, dtype=None,
                       schema=None):
    """
    Retrieves defined inputs in already declared variables
    bsed on their names.

    @param      input_names     input names
    @param      variables       registered variables created
                                by previous operators
    @param      dtype           float computational type
    @param      schema          defined inputs by schema (*expected_inputs*)
    @return                     typed inputs
                                as ``tuple(name, type)``
    """
    def guess_type_variable(name, schema):
        if variables is None:
            if (schema is None or
                    not isinstance(schema, (DataType, tuple))):
                return (  # pragma: no cover
                    DoubleTensorType() if dtype == numpy.float64 else FloatTensorType())
            return schema if isinstance(schema, DataType) else schema[1]
        if name in variables:
            ty = variables[name]
            if isinstance(ty, DataType):
                shape = ty.shape
                if 0 in shape:
                    raise RuntimeError(  # pragma: no cover
                        "Shape cannot be empty: name='{}', var={}".format(
                            name, ty))
                return variables[name]
            if isinstance(ty, dict) and 'value' in ty:
                # constant
                arr = ty['value']
                try:
                    return _guess_type(arr)
                except RuntimeError as e:  # pragma: no cover
                    raise RuntimeError(
                        "Unable to guess type of variable '{}' - {}."
                        "".format(name, arr)) from e
            raise NotImplementedError(  # pragma: no cover
                "Unable to guess type for '{}' form '{}'.".format(
                    name, variables[name]))
        if isinstance(schema, (DataType, tuple)):
            sch = schema if isinstance(schema, DataType) else schema[1]
            if not isinstance(sch, str):
                return sch
        # Inputs. Let's assume it is a vector of floats.
        return DoubleTensorType() if dtype == numpy.float64 else FloatTensorType()

    if schema is None or len(schema) < len(input_names):
        inputs = [(name, guess_type_variable(name, None))
                  for name in input_names]
    else:
        inputs = [(name, guess_type_variable(name, schema=sch))
                  for name, sch in zip(input_names, schema)]
    return inputs


def get_defined_outputs(outputs, onnx_node, typed_inputs=None, variables=None,
                        dtype=None, schema=None, schema_inputs=None):
    """
    Gets types of predefined outputs when they cannot be inferred.
    Some part of it should be automated based
    on type constraints.

    :param outputs: requested outputs
    :param onnx_node: :epkg:`ONNX` node definition
    :param typed_inputs: known typed inputs of the node as `tuple(name, type)`
    :param variables: registered variables created by previous operators
    :param dtype: float computational type
    :param schema: defined outputs by schema (*expected_outputs*)
    :param schema_inputs: defined inputs by schema (*expected_inputs*)
    :return: typed outputs as ``tuple(name, type)``
    """
    if schema is None:
        ft = DoubleTensorType if dtype == numpy.float64 else FloatTensorType
    elif len(schema) != 1:
        raise ValueError(
            "schema should only contain one output not {}.".format(schema))
    else:
        if isinstance(schema, DataType):
            ft = schema[0].__class__
        else:
            ft = schema[0][1].__class__

    if onnx_node.op_type in {'ZipMap', 'ArgMin', 'ArgMax', 'Shape',
                             'Greater', 'Less', 'Equal', 'TopK',
                             'Cast', 'ArrayFeatureExtractor',
                             'Reshape', 'Transpose', 'Scan',
                             'ConstantOfShape'}:
        if onnx_node.op_type == "ZipMap":
            # ZipMap
            otype = SequenceType(DictionaryType(
                Int64Type(), ft()))
            outputs = [(name, otype) for name in outputs]
        elif (onnx_node.op_type in ("ArgMin", "ArgMax", 'Shape') and
                len(outputs) == 1):
            # ArgMin, ArgMax, Shape
            outputs = [(outputs[0], Int64TensorType())]
        elif (onnx_node.op_type in ("Greater", "Less", 'Equal') and
                len(outputs) == 1):
            # Greater, Less, Equal
            outputs = [(outputs[0], BooleanTensorType())]
        elif onnx_node.op_type == "TopK" and len(outputs) == 2:
            # TopK
            if len(typed_inputs) != 2:
                raise RuntimeError(  # pragma: no cover
                    "Wrong typed_inputs, got {}.".format(typed_inputs))
            outputs = [(outputs[0], typed_inputs[0][1]),
                       (outputs[1], Int64TensorType())]
        elif onnx_node.op_type == "Cast" and len(outputs) == 1:
            # Cast
            ttyp = _guess_type_proto(onnx_node.attribute[0].i, dims=None)
            outputs = [(outputs[0], ttyp)]
        elif onnx_node.op_type == "ArrayFeatureExtractor":
            # ArrayFeatureExtractor
            if len(typed_inputs) != 2:
                raise RuntimeError(  # pragma: no cover
                    "Wrong typed_inputs, got {}.".format(typed_inputs))
            outputs = [(outputs[0], typed_inputs[0][1])]
        elif onnx_node.op_type in ('Reshape', 'Transpose'):
            # Reshape
            outputs = [(outputs[0], typed_inputs[0][1].__class__())]
        elif onnx_node.op_type == 'Scan':
            # Scan
            if len(outputs) != len(typed_inputs):
                raise RuntimeError(  # pragma: no cover
                    "Dimension mismatch, operator Scan should have "
                    "the same number of inputs and outputs {} != {}"
                    ".".format(len(outputs), len(typed_inputs)))
            outputs = [(o, t[1].__class__())
                       for o, t in zip(outputs, typed_inputs)]
        elif onnx_node.op_type == "ConstantOfShape":
            # ConstantOfShape
            outputs = [(outputs[0], ft())]
    elif 'Classifier' in onnx_node.op_type:
        # Good chance that's a classifier.
        outputs = [(outputs[0], Int64TensorType()),
                   (outputs[1], ft())]
    else:
        if schema_inputs is not None and schema is not None:
            dt = {}
            for got, exp in zip(typed_inputs, schema_inputs):
                if isinstance(exp[1], str):
                    dt[exp[1]] = got
            out = []
            for i in range(len(outputs)):  # pylint: disable=C0200
                o = outputs[i]
                if isinstance(o, str):
                    exp = schema[i]
                    if exp[1] in dt:
                        out.append((o, dt[exp[1]][1].__class__()))
                    else:
                        nt = _guess_type_proto_str(exp[1], None)
                        out.append((o, nt))
                elif (isinstance(o, tuple) and
                        (isinstance(o[1], str) or o[1] is None)):
                    exp = schema[i]
                    if exp[1] in dt:
                        out.append((o[0], dt[exp[1]][1].__class__()))
                    else:
                        nt = _guess_type_proto_str(exp[1], None)
                        out.append((o[0], nt))
                else:
                    out.append(o)
            outputs = out
        elif len(typed_inputs) == 1 and len(outputs) == 1:
            # Default case
            # Assuming the only output is the same as the only input.
            outputs = [(outputs[0], typed_inputs[0][1])]
        else:
            # Default
            outputs = [(name, ft()) for name in outputs]

    for name, typ in outputs:
        if typ in ('T', None, '', 'I'):
            raise NotImplementedError(  # pragma: no cover
                "Undefined output type: %r (outputs=%r, typed_inputs=%r, "
                "dtype=%r, schema=%r, schema_inputs=%r, onnx_node=%r, "
                "variables=%r)." % (
                    typ, outputs, typed_inputs, dtype,
                    schema, schema_inputs, onnx_node, variables))
        if not isinstance(name, str):
            raise NotImplementedError(  # pragma: no cover
                "Undefined output type: %r (outputs=%r, typed_inputs=%r, "
                "dtype=%r, schema=%r, schema_inputs=%r, onnx_node=%r, "
                "variables=%r)." % (
                    typ, outputs, typed_inputs, dtype,
                    schema, schema_inputs, onnx_node, variables))
    return outputs


def proto2vars(values):
    """
    Converts proto values to Variables.
    """
    def ptype2vttype(it, shape):
        if it == TensorProto.FLOAT:  # pylint: disable=E1101
            return FloatTensorType(shape)
        if it == TensorProto.DOUBLE:  # pylint: disable=E1101
            return DoubleTensorType(shape)
        if it == TensorProto.INT64:  # pylint: disable=E1101
            return Int64TensorType(shape)
        if it == TensorProto.INT32:  # pylint: disable=E1101
            return Int32TensorType(shape)
        if it == TensorProto.BOOL:  # pylint: disable=E1101
            return BooleanTensorType(shape)
        if it == TensorProto.STRING:  # pylint: disable=E1101
            return StringTensorType(shape)
        if Float16TensorType is None:
            if it == TensorProto.FLOAT16:  # pylint: disable=E1101
                return Float16TensorType(shape)
        raise NotImplementedError(  # pragma: no cover
            "Unrecognized proto type {} with shape {}".format(it, shape))

    def ptype2vtype(it):
        if it == TensorProto.FLOAT:  # pylint: disable=E1101
            return FloatType()
        if it == TensorProto.INT64:  # pylint: disable=E1101
            return Int64Type()
        raise NotImplementedError(  # pragma: no cover
            "Unrecognized proto type {}".format(it))

    res = []
    for v_ in values:
        v = v_
        name = v.name if hasattr(v, 'name') else None
        if hasattr(v, 'type') and str(v.type) != '':
            t = v.type
            v = proto2vars([t])[0][1]
        elif hasattr(v, 'sequence_type') and str(v.sequence_type) != '':
            subtype = proto2vars([v.sequence_type.elem_type])[0][1]
            v = SequenceType(subtype)
        elif hasattr(v, 'tensor_type') and str(v.tensor_type) != '':
            tt = v.tensor_type
            el = tt.elem_type
            shape = tt.shape
            dim = shape.dim
            if len(dim) == 0:
                shape = []
            else:
                shape = [dim[i].dim_value for i in range(len(dim))]
            v = ptype2vttype(el, shape)
        elif hasattr(v, 'map_type') and str(v.map_type) != '':
            mt = v.map_type
            keyt = ptype2vtype(mt.key_type)
            valt = proto2vars([mt.value_type])[0][1]
            v = DictionaryType(keyt, valt)
        else:
            raise RuntimeError(  # pragma: no cover
                "Unable to build a variable from {}.".format(v))
        if v.shape is not None and 0 in v.shape:
            # Replaces 0 by None
            new_shape = tuple(None if d == 0 else d for d in v.shape)
            if new_shape in ((None, ), None):
                v = v.__class__()
            else:
                v = v.__class__(new_shape)
        if v.shape is not None and 0 in v.shape:
            raise RuntimeError(  # pragma: no cover
                "Shape cannot be empty: '{}': {}.".format(
                    name, v_))
        res.append((name, v))
    return res



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *ops_cpu*.
"""
from onnx.defs import onnx_opset_version
from ...tools.asv_options_helper import benchmark_version
from ._op import OpRunCustom
from ._op_list import __dict__ as d_op_list


_additional_ops = {}


def register_operator(cls, name=None, overwrite=True):
    """
    Registers a new runtime operator.

    @param      cls         class
    @param      name        by default ``cls.__name__``,
                            or *name* if defined
    @param      overwrite   overwrite or raise an exception
    """
    if name is None:
        name = cls.__name__
    if name not in _additional_ops:
        _additional_ops[name] = cls
    elif not overwrite:
        raise RuntimeError(  # pragma: no cover
            "Unable to overwrite existing operator '{}': {} "
            "by {}".format(name, _additional_ops[name], cls))


def get_opset_number_from_onnx(benchmark=False):
    """
    Retuns the current :epkg:`onnx` opset
    based on the installed version of :epkg:`onnx`.

    @param      benchmark       returns the latest
                                version usable for benchmark
    @eturn                      opset number
    """
    if benchmark:
        return benchmark_version()[-1]
    return onnx_opset_version()


def load_op(onnx_node, desc=None, options=None):
    """
    Gets the operator related to the *onnx* node.

    @param      onnx_node       :epkg:`onnx` node
    @param      desc            internal representation
    @param      options         runtime options
    @return                     runtime class
    """
    if desc is None:
        raise ValueError("desc should not be None.")  # pragma no cover
    name = onnx_node.op_type
    opset = options.get('target_opset', None) if options is not None else None
    current_opset = get_opset_number_from_onnx()
    chosen_opset = current_opset
    if opset == current_opset:
        opset = None
    if opset is not None:
        if not isinstance(opset, int):
            raise TypeError(  # pragma no cover
                "opset must be an integer not {}".format(type(opset)))
        name_opset = name + "_" + str(opset)
        for op in range(opset, 0, -1):
            nop = name + "_" + str(op)
            if nop in d_op_list:
                name_opset = nop
                chosen_opset = op
                break
    else:
        name_opset = name

    if name_opset in _additional_ops:
        cl = _additional_ops[name_opset]
    elif name in _additional_ops:
        cl = _additional_ops[name]
    elif name_opset in d_op_list:
        cl = d_op_list[name_opset]
    elif name in d_op_list:
        cl = d_op_list[name]
    else:
        raise NotImplementedError(  # pragma no cover
            "Operator '{}' has no runtime yet. Available list:\n"
            "{}\n--- +\n{}".format(
                name, "\n".join(sorted(_additional_ops)),
                "\n".join(
                    _ for _ in sorted(d_op_list)
                    if "_" not in _ and _ not in {'cl', 'clo', 'name'})))

    if hasattr(cl, 'version_higher_than'):
        opv = min(current_opset, chosen_opset)
        if cl.version_higher_than > opv:
            # The chosen implementation does not support
            # the opset version, we need to downgrade it.
            if ('target_opset' in options and
                    options['target_opset'] is not None):  # pragma: no cover
                raise RuntimeError(
                    "Supported version {} > {} (opset={}) required version, "
                    "unable to find an implementation version {} found "
                    "'{}'\n--ONNX--\n{}\n--AVAILABLE--\n{}".format(
                        cl.version_higher_than, opv, opset,
                        options['target_opset'], cl.__name__, onnx_node,
                        "\n".join(
                            _ for _ in sorted(d_op_list)
                            if "_" not in _ and _ not in {'cl', 'clo', 'name'})))
            options = options.copy()
            options['target_opset'] = current_opset
            return load_op(onnx_node, desc=desc, options=options)

    if options is None:
        options = {}  # pragma: no cover
    return cl(onnx_node, desc=desc, **options)



"""
@file
@brief Initializes a quantized tensor from float values.
"""
import numpy
from skl2onnx.algebra.onnx_ops import OnnxQLinearConv  # pylint: disable=E0611
from ...onnxrt import OnnxInference


class QuantizedTensor:
    """
    Instantiates a quantized tensor (uint8) from a float tensor.

    :param data: array
    :param scale: scale if data.dtype is float32, None otherwise
    :param zero_point: zero_point if data.dtype is float32, None otherwise
    """

    def __init__(self, data, scale=None, zero_point=None):
        "constructor"
        if data.dtype == numpy.float32:
            if scale is not None or zero_point is not None:
                raise ValueError("scale and zero_point are ignored.")
            self._init(data)
        elif data.dtype == numpy.uint8:
            if scale is None or zero_point is None:
                raise ValueError("scale and zero_point must be specified.")
            self.quantized_ = data
            self.scale_ = numpy.float32(scale)
            self.zero_point_ = numpy.uint8(zero_point)

    def _init(self, data):
        "Initialization when dtype is float32."
        rav = data.flatten().astype(numpy.float32)
        mini = min(rav.min(), numpy.float32(0))
        maxi = max(rav.max(), numpy.float32(0))

        info = numpy.iinfo(numpy.uint8)
        qmin = numpy.float32(info.min)
        qmax = numpy.float32(info.max)

        self.scale_ = (maxi - mini) / (qmax - qmin)
        initial_zero_point = qmin - mini / self.scale_
        self.zero_point_ = numpy.uint8(numpy.round(
            max(qmin, min(qmax, initial_zero_point))))

        self.quantized_ = numpy.empty(data.size, dtype=numpy.uint8)
        for i in range(0, data.size):
            clamped_val = numpy.float32(
                max(qmin, min(qmax, numpy.round(data[i] / self.scale_) + self.zero_point_)))
            self.quantized_[i] = numpy.uint8(clamped_val)

        if self.quantized_.dtype != numpy.uint8:
            raise TypeError(
                "dtype={} not uint8".format(self.quantized_.dtype))


class QuantizedBiasTensor:
    """
    Instantiates a quantized tensor (uint8) with bias
    from a float tensor.

    :param data: array
    :param X_or_scale: a @see cl QuantizedTensor or a float
    :param zero_point: a @see cl QuantizedTensor or or None
    """

    def __init__(self, data, X_or_scale, W: QuantizedTensor = None):
        if W is None:
            self.quantized_ = data
            self.scale_ = numpy.float32(X_or_scale)
        else:
            self.scale_ = X_or_scale.scale_ * W.scale_

            self.quantized_ = numpy.empty(data.size(), dtype=numpy.int32)
            for i in range(0, data.size()):
                self.quantized_[i] = numpy.int32(
                    numpy.floor(data[i] / (X_or_scale.scale_ * W.scale_)))
        if self.quantized_.dtype != numpy.int32:
            raise TypeError(
                "dtype={} not int32".format(self.quantized_.dtype))


def test_qlinear_conv(x: QuantizedTensor, x_shape,
                      w: QuantizedTensor, w_shape,
                      b: QuantizedBiasTensor,
                      y: QuantizedTensor, y_shape,
                      opset=None, runtime='python',
                      pads=None, strides=None, group=None):
    """
    Checks a runtime for operator `QLinearConv`.

    :param x: @see cl QuantizedTensor
    :param x_shape: shape of X
    :param w: @see cl QuantizedTensor
    :param w_shape: shape of W
    :param b: @see cl QuantizedBiasTensor or None
    :param y: expected output, @see cl QuantizedTensor or None
    :param y_shape: shape of Y
    :param opset: desired onnx opset
    :param runtime: runtime for @see cl OnnxInference
    :param pads: optional parameter for operator `QLinearConv`
    :param strides: optional parameter for operator `QLinearConv`
    :param group: optional paramerer for operator `QLinearConv`
    """
    if opset is None:
        from ...tools.asv_options_helper import get_opset_number_from_onnx
        opset = get_opset_number_from_onnx()

    kwargs = {}
    if pads is not None:
        kwargs['pads'] = pads
    if strides is not None:
        kwargs['strides'] = strides
    if group is not None:
        kwargs['group'] = group

    if b is None:
        inputs_list = [
            'x', 'x_scale', 'x_zero_point', 'w', 'w_scale', 'w_zero_point',
            'y_scale', 'y_zero_point']
        inputs = {'x': x.quantized_.reshape(x_shape),
                  'x_scale': x.scale_, 'x_zero_point': x.zero_point_,
                  'w': w.quantized_.reshape(w_shape),
                  'w_scale': w.scale_, 'w_zero_point': w.zero_point_,
                  'y_scale': y.scale_, 'y_zero_point': y.zero_point_}
    else:
        inputs_list = [
            'x', 'x_scale', 'x_zero_point', 'w', 'w_scale', 'w_zero_point',
            'y_scale', 'y_zero_point', 'b']
        inputs = {'x': x.quantized_.reshape(x_shape),
                  'x_scale': x.scale_, 'x_zero_point': x.zero_point_,
                  'w': w.quantized_.reshape(w_shape),
                  'w_scale': w.scale_, 'w_zero_point': w.zero_point_,
                  'y_scale': y.scale_, 'y_zero_point': y.zero_point_,
                  'b': b.quantized_}

    for k in inputs:  # pylint: disable=C0206
        v = inputs[k]
        if len(v.shape) == 0:
            inputs[k] = numpy.array([v], dtype=v.dtype)

    node = OnnxQLinearConv(*inputs_list, output_names=['y'],
                           op_version=opset, **kwargs)
    model_def = node.to_onnx(inputs, target_opset=opset)

    oinf = OnnxInference(model_def, runtime=runtime)
    got = oinf.run(inputs)['y']
    expected = y.quantized_.reshape(y_shape)
    if got.dtype != expected.dtype:
        raise TypeError(
            "Unexpected output dtype:\nEXPECTED\n{}\nGOT\n{}"
            "".format(expected, got))
    diff = numpy.abs(got.ravel().astype(numpy.float32) -
                     expected.ravel().astype(numpy.float32))
    mdiff = diff.max()
    if mdiff > 1e-5:
        raise ValueError(
            "Unexpected output maximum difference={}:\nEXPECTED\n{}\nGOT\n{}"
            "".format(mdiff, expected, got))



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *ops_onnxruntime*.
"""
import numpy
import onnx.defs
from onnx.helper import make_tensor
import skl2onnx.algebra.onnx_ops as alg
try:
    import skl2onnx.algebra.custom_ops as alg2
except ImportError:  # pragma: no cover
    # older version of skl2onnx
    alg2 = alg
from ...tools.ort_wrapper import (
    InferenceSession, SessionOptions, RunOptions,
    GraphOptimizationLevel, OrtInvalidArgument,
    OrtNotImplemented, OrtInvalidGraph, OrtFail)
from ...onnx_tools.onnx2py_helper import guess_proto_dtype
from ...onnx_tools.optim.graph_schema_helper import (
    get_defined_inputs, get_defined_outputs, proto2vars)


_schemas = {
    schema.name: schema for schema in onnx.defs.get_all_schemas_with_history()}


class OpRunOnnxRuntime:
    """
    Unique operator which calls :epkg:`onnxruntime`
    to compute predictions for one operator.
    """

    def __init__(self, onnx_node, desc=None, variables=None,
                 dtype=None, **options):
        """
        @param      onnx_node               :epkg:`onnx` node
        @param      desc                    internal representation
        @param      variables               registered variables created by previous operators
        @param      dtype                   float computation type
        @param      options                 runtime options
        """
        self._provider = 'onnxruntime'
        self.onnx_node = onnx_node
        self.desc = desc
        self._schema = _schemas.get(onnx_node.op_type, None)
        if desc is not None:
            if 'atts' in desc:
                for a, b in desc['atts'].items():
                    if not isinstance(b, dict) or 'value' not in b:
                        raise ValueError(  # pragma: no cover
                            "Unexpected value {}.".format(b))
                    options[a] = b['value']

        self.options = options
        self.dtype = dtype
        self._init(variables)

    def _name_mapping(self, inputs):
        mapping = {}
        new_inputs = []
        for name in inputs:
            if name in mapping:
                i = 0
                new_name = "{}_{}".format(name, i)
                while new_name in mapping:
                    i += 1  # pragma: no cover
                    new_name = "{}_{}".format(name, i)  # pragma: no cover
                mapping[new_name] = name
                new_inputs.append(new_name)
            else:
                new_inputs.append(name)
                mapping[name] = name
        return mapping, new_inputs

    def _guess_proto_type(self, dtype):
        return guess_proto_dtype(dtype)

    def _init(self, variables=None):
        """
        Initializes the node.

        :param variables: registered variables created by previous operators

        The current implementation for operator *Scan*
        only works for matrices.
        """
        custom_nodes = self.options.get('nodes', None)
        if (custom_nodes is not None and
                self.onnx_node.op_type in custom_nodes):
            self.alg_class = custom_nodes[self.onnx_node.op_type]
        else:
            try:
                self.alg_class = getattr(alg2, 'Onnx' + self.onnx_node.op_type)
            except AttributeError:
                self.alg_class = getattr(alg, 'Onnx' + self.onnx_node.op_type)

        inputs = list(self.onnx_node.input)
        self.mapping, self.inputs = self._name_mapping(inputs)
        self.outputs = list(self.onnx_node.output)

        options = self.options.copy()
        options.pop('nodes', None)
        target_opset = options.pop('target_opset', None)
        domain = options.pop('domain', None)
        disable_optimisation = options.pop('disable_optimisation', False)
        session_options = options.pop('session_options', False)
        ir_version = options.pop('ir_version', None)

        if domain == '' and target_opset < 9:
            # target_opset should be >= 9 not {} for main domain.
            # We assume it was the case when the graph was created.
            pass

        if self.onnx_node.op_type == 'ConstantOfShape':
            for k in options:
                v = options[k]
                if isinstance(v, numpy.ndarray):
                    options[k] = make_tensor(
                        k, self._guess_proto_type(v.dtype),
                        v.shape, v.tolist())

            self.inst_ = self.alg_class(*self.inputs, output_names=self.outputs,
                                        op_version=target_opset, **options)
            inputs = get_defined_inputs(
                self.inputs, variables, dtype=self.dtype)
            try:
                self.onnx_ = self.inst_.to_onnx(inputs, target_opset=target_opset,
                                                domain=domain)
                if "dim_value: 0" in str(self.onnx_):
                    raise RuntimeError(  # pragma: no cover
                        "Probable issue as one dimension is null.\n--\n{}".format(
                            self.onnx_))
            except AttributeError as e:  # pragma: no cover
                # older version of skl2onnx
                self.onnx_ = self.inst_.to_onnx(inputs)
                if "dim_value: 0" in str(self.onnx_):
                    raise RuntimeError(
                        "Probable issue as one dimension is null.\n--\n{}".format(
                            self.onnx_)) from e
            forced = False
        elif self.onnx_node.op_type == 'Scan':
            self.inst_ = self.alg_class(
                *self.inputs, output_names=self.outputs,
                op_version=target_opset, **options)
            inputs = get_defined_inputs(
                self.inputs, variables, dtype=self.dtype)
            outputs = get_defined_outputs(
                self.outputs, self.onnx_node, inputs, variables,
                dtype=self.dtype)
            inputs = [(name, cl.__class__([None, None]))
                      for (name, cl) in inputs]
            outputs = [(name, cl.__class__([None, None]))
                       for (name, cl) in outputs]
            self.onnx_ = self.inst_.to_onnx(inputs, outputs=outputs,
                                            target_opset=target_opset,
                                            domain=domain)
            if "dim_value: 0" in str(self.onnx_):
                raise RuntimeError(  # pragma: no cover
                    "Probable issue as one dimension is null.\n--\n{}".format(
                        self.onnx_))
            forced = True
        else:
            self.inst_ = self.alg_class(*self.inputs, output_names=self.outputs,
                                        op_version=target_opset, domain=domain,
                                        **options)
            inputs = get_defined_inputs(
                self.inputs, variables, dtype=self.dtype,
                schema=self.alg_class.expected_inputs)

            try:
                self.onnx_ = self.inst_.to_onnx(
                    inputs, target_opset=target_opset, domain=domain)
                if "dim_value: 0" in str(self.onnx_):
                    raise RuntimeError(  # pragma: no cover
                        "Probable issue as one dimension is null.\n--\n{}\n---\n{}".format(
                            self.onnx_, inputs))
                forced = False
            except (RuntimeError, ValueError):
                # Let's try again by forcing output types.
                forced = True
                outputs = get_defined_outputs(
                    self.outputs, self.onnx_node, inputs, variables,
                    dtype=self.dtype, schema=self.alg_class.expected_outputs,
                    schema_inputs=self.alg_class.expected_inputs)
                try:
                    self.onnx_ = self.inst_.to_onnx(inputs, outputs=outputs,
                                                    target_opset=target_opset,
                                                    domain=domain)
                except NotImplementedError as e:
                    raise NotImplementedError(
                        "Unable to instantiate node {} inputs={} "
                        "self.inputs={} outputs={} variables={} "
                        "dtype={}".format(
                            self.alg_class, inputs, self.inputs,
                            outputs, variables, self.dtype)) from e
                if "dim_value: 0" in str(self.onnx_):
                    raise RuntimeError(  # pragma: no cover
                        "Probable issue as one dimension is null.\n--\n{}".format(
                            self.onnx_)) from e

        if len(self.onnx_.graph.output) != len(self.outputs):  # pragma: no cover
            # Something is wrong, falls back to default plan.
            forced = True
            outputs = get_defined_outputs(
                self.outputs, self.onnx_node, inputs, variables,
                dtype=self.dtype, schema=self.alg_class.expected_outputs)
            self.onnx_ = self.inst_.to_onnx(inputs, outputs=outputs,
                                            target_opset=target_opset,
                                            domain=domain)
            if "dim_value: 0" in str(self.onnx_):
                raise RuntimeError(  # pragma: no cover
                    "Probable issue as one dimension is null.\n--\n{}".format(
                        self.onnx_))
        else:
            lo = list(self.onnx_.graph.output)
            outputs = proto2vars(lo)

        sess_options = session_options or SessionOptions()
        self.run_options = RunOptions()

        if session_options is None:
            try:
                sess_options.session_log_severity_level = 3
                # sess_options.sessions_log_verbosity_level = 0
            except AttributeError:
                # onnxruntime not recent enough.
                pass
            try:
                self.run_options.run_log_severity_level = 3
                # self.run_options.run_log_verbosity_level = 0
            except AttributeError:
                # onnxruntime not recent enough.
                pass
            if disable_optimisation:
                sess_options.graph_optimization_level = (  # pragma: no cover
                    GraphOptimizationLevel.ORT_DISABLE_ALL)
        elif disable_optimisation:
            raise RuntimeError(  # pragma: no cover
                "session_options and disable_optimisation cannot be defined "
                "at the same time.")

        if ir_version is not None:
            self.onnx_.ir_version = ir_version
        try:
            self.sess_ = InferenceSession(
                self.onnx_.SerializeToString(), sess_options=sess_options)
        except (RuntimeError, OrtNotImplemented, OrtInvalidGraph, OrtFail) as e:
            raise RuntimeError(
                "Unable to load node '{}' (output type was {}) inputs={} "
                "self.inputs={} self.onnx_node.input={} "
                "variables={} mapping={} "
                "expected_inputs={}\n{}".format(
                    self.onnx_node.op_type,
                    "guessed" if forced else "inferred",
                    inputs, self.inputs, self.onnx_node.input,
                    variables, self.mapping,
                    self.alg_class.expected_inputs,
                    self.onnx_)) from e
        self.typed_outputs_ = outputs

    def run(self, *args, **kwargs):
        """
        Should be overwritten.
        """
        inputs = {name: val for name, val in zip(self.inputs, args)}

        try:
            res = self.sess_.run(None, inputs, self.run_options)
        except (RuntimeError, OrtInvalidArgument) as e:  # pragma: no cover
            dtypes = {k: v.dtype for k, v in inputs.items()}
            shapes = {k: v.shape for k, v in inputs.items()}
            exp = [_.name for _ in self.sess_.get_inputs()]
            exp_types = [_.type for _ in self.sess_.get_inputs()]
            raise RuntimeError(
                "Predictions failed. List of inputs: {}, class={}"
                "\ndtypes={}\nshapes={}\nexpected={}\nexpected={}\n"
                "exception={}\n--ONNX--\n{}".format(
                    list(sorted(inputs)), self.alg_class, dtypes,
                    shapes, exp, exp_types, e, self.onnx_)) from e
        return tuple(res)



"""
@file
@brief Functions used to predict the cost of a transposition.
"""
import numpy


_ml_transpose_coefs = {
    'CST_': 0.4720163707200312,
    'begin': 0.0,
    'dbegin': 0.0,
    'dend': 0.0,
    'dim': 0.0,
    'discont': 0.0180766756730043,
    'edit': 0.06940318842803926,
    'end': 0.0,
    'end16': 0.0,
    'end32': 0.0,
    'ibegin16': 0.0,
    'ibegin2': 0.0,
    'ibegin32': 0.0,
    'ibegin4': 0.0,
    'ibegin64': 0.0,
    'ibegin8': 0.04389296884016416,
    'iend16': 0.5316238365817172,
    'iend2': 0.16287259236456927,
    'iend32': 0.0,
    'iend4': 0.0,
    'iend64': 0.0,
    'iend8': 0.0,
    'middle': 1.3381940773605624e-06,
    'rbegin': 0.0,
    'rdiscont': 0.0,
    'redit': 0.18604684802855143,
    'rend': 0.0,
    'rend16': 0.0,
    'rend32': 0.0,
    'rev': 0.42909943168149206,
    'rmiddle': 0.0,
    'rot': 0.22272566615803094,
    'size': 2.8663794075460607e-06}


def _edit_distance(mot1, mot2):
    dist = {(-1, -1): 0}
    if len(mot1) == 0:
        for j, d in enumerate(mot2):
            dist[-1, j] = dist[-1, j - 1] + 1
            dist[j, -1] = dist[j - 1, -1] + 1
    for i, c in enumerate(mot1):
        dist[i, -1] = dist[i - 1, -1] + 1
        dist[-1, i] = dist[-1, i - 1] + 1
        for j, d in enumerate(mot2):
            opt = []
            if (i - 1, j) in dist:
                x = dist[i - 1, j] + 1
                opt.append((x, (i - 1, j)))
            if (i, j - 1) in dist:
                x = dist[i, j - 1] + 1
                opt.append((x, (i, j - 1)))
            if (i - 1, j - 1) in dist:
                x = dist[i - 1, j - 1] + (1 if c != d else 0)
                opt.append((x, (i - 1, j - 1)))
            mi = min(opt)
            dist[i, j] = mi[0]

    return dist[len(mot1) - 1, len(mot2) - 1]


def _is_rotation(perm):
    t = tuple(perm)
    c = list(range(len(perm)))
    for i in range(len(c)):
        for k in range(len(c)):  # pylint: disable=C0200
            c[k] = (k + i) % len(c)
        if t == tuple(c):
            return True
    return False


def _relu(x, origin=0):
    return origin if x < origin else x


def compute_transposition_features(shape, perm):
    """
    Given a shape and a permutation, computes many features
    used to predict the cost of the transposition.

    :param shape: shape
    :param perm: permutation
    :return: dictionary of features

    .. runpython::
        :showcode:

        import pprint
        from mlprodict.testing.einsum.einsum_ml import (
            compute_transposition_features)

        pprint.pprint(
            compute_transposition_features((3, 5, 7), (2, 1, 0)))
    """
    total = numpy.prod(numpy.array(shape, dtype=numpy.int64))

    begin = 1
    dbegin = 0
    for i, p in enumerate(perm):
        if p != i:
            break
        dbegin += 1
        begin *= shape[i]

    end = 1
    dend = 0
    for i in range(len(perm) - 1, -1, -1):
        if perm[i] != i:
            break
        dend += 1
        end *= shape[i]

    dis_cont = 0
    for i in range(1, len(shape)):
        if perm[i] != perm[i - 1] + 1:
            dis_cont += 1

    middle = max(1, int(total / (end * begin)))
    feat = dict(size=total, begin=begin, end=end, middle=middle,
                dim=len(shape), discont=dis_cont)

    for c in [16, 32]:
        feat["end%d" % c] = _relu(end, c)

    keys = list(feat)
    for k in keys:
        if k in {'dim', 'cpu', 'size'}:
            continue
        feat['r%s' % k] = float(feat[k] / total)

    for c in [2, 4, 8, 16, 32, 64]:
        feat["iend%d" % c] = float(end >= c)
        feat["ibegin%d" % c] = float(begin >= c)

    # feat['CST'] = 1
    feat['CST_'] = -1
    feat['dbegin'] = - dbegin
    feat['dend'] = - dend

    keys = list(feat)
    for k in keys:
        if k.startswith('end') or k.startswith('begin'):
            feat[k] = - feat[k]
        elif k.startswith('rend') or k.startswith('rbegin'):
            feat[k] = - feat[k]
        elif k.startswith('iend') or k.startswith('ibegin'):
            feat[k] = - feat[k]
        elif k == "rdiscont":
            feat[k] = - feat[k]

    idp = list(range(len(perm)))
    feat["rot"] = -1 if _is_rotation(perm) else 0
    feat["rev"] = 1 if perm == tuple(idp[::-1]) else 0
    feat["edit"] = _edit_distance(idp, perm)
    feat["redit"] = feat["edit"] / len(idp)
    return feat


def predict_transposition_cost(shape, perm, coefs=None):
    """
    Given a shape and a permutation, predicts the cost of the
    transposition.

    :param shape: shape
    :param perm: permutation
    :param coefs: trained coefficients or None to get
        the default ones
    :return: dictionary of features

    .. runpython::
        :showcode:

        import pprint
        from mlprodict.testing.einsum.einsum_ml import (
            compute_transposition_features)

        pprint.pprint(
            compute_transposition_features((3, 5, 7), (2, 1, 0)))
    """
    if coefs is None:
        coefs = _ml_transpose_coefs
    feat = compute_transposition_features(shape, perm)
    res = 0
    for k, v in feat.items():
        res += v * coefs[k]
    return max(0., res / 1000)



"""
@file
@brief :epkg:`numpy` functions implemented with :epkg:`onnx`.

.. versionadded:: 0.6
"""
import numpy
from onnx import onnx_pb as onnx_proto  # pylint: disable=E1101
from onnx.helper import make_tensor
from skl2onnx.algebra.onnx_ops import (  # pylint: disable=E0611
    OnnxAbs,
    OnnxAcos, OnnxAcosh,
    OnnxAdd,
    OnnxArgMax,
    OnnxArgMin,
    OnnxAsin, OnnxAsinh,
    OnnxAtan, OnnxAtanh,
    OnnxCeil,
    OnnxClip,
    OnnxCompress, OnnxConcat,
    OnnxConstantOfShape,
    OnnxCos, OnnxCosh,
    OnnxCumSum,
    OnnxDet,
    OnnxEinsum,
    OnnxErf,
    OnnxExp,
    OnnxFloor,
    OnnxIdentity, OnnxIsNaN,
    OnnxLog,
    OnnxMatMul,
    OnnxPad,
    OnnxReciprocal,
    OnnxReduceMax,
    OnnxReduceMean,
    OnnxReduceMin,
    OnnxReduceProd,
    OnnxReduceSum,
    OnnxRelu,
    OnnxRound,
    OnnxSigmoid,
    OnnxSign,
    OnnxSin, OnnxSinh,
    OnnxSqrt,
    OnnxSqueeze,
    OnnxTan, OnnxTanh, OnnxTopK, OnnxTranspose,
    OnnxUnsqueeze,
    OnnxWhere)
from .onnx_variable import OnnxVar, MultiOnnxVar as xtuple


def abs(x):
    "See :epkg:`numpy:abs`."
    return OnnxVar(x, op=OnnxAbs)


def acos(x):
    "See :epkg:`numpy:acos`."
    return OnnxVar(x, op=OnnxAcos)


def acosh(x):
    "See :epkg:`numpy:acosh`."
    return OnnxVar(x, op=OnnxAcosh)


def amax(x, axis=None, keepdims=0):
    "See :epkg:`numpy:amax`."
    if axis is None:
        return OnnxVar(x, op=OnnxReduceMax, keepdims=keepdims)
    if not isinstance(axis, list):
        axis = [axis]
    return OnnxVar(x, op=OnnxReduceMax, keepdims=keepdims, axes=axis)


def amin(x, axis=None, keepdims=0):
    "See :epkg:`numpy:amin`."
    if axis is None:
        return OnnxVar(x, op=OnnxReduceMin, keepdims=keepdims)
    if not isinstance(axis, list):
        axis = [axis]
    return OnnxVar(x, op=OnnxReduceMin, keepdims=keepdims, axes=axis)


def arange(start, stop, step=1):
    "See :epkg:`numpy:arange`, *start*, *stop* must be specified."
    if step != 1:
        raise NotImplementedError(  # pragma: no cover
            "The function is not implemented for step != 1 (step=%r)." % step)
    if isinstance(start, (int, numpy.int64)):
        start = numpy.array([start], dtype=numpy.int64)
    if isinstance(stop, (int, numpy.int64)):
        stop = numpy.array([stop], dtype=numpy.int64)
    value = make_tensor(
        "value", onnx_proto.TensorProto.INT64, (1, ), [step])  # pylint: disable=E1101
    _cst = OnnxVar(stop - start, op=OnnxConstantOfShape, value=value)
    cs = OnnxVar(_cst,
                 numpy.array([0], dtype=numpy.int64),
                 op=OnnxCumSum)
    diff = start - numpy.array([step], dtype=numpy.int64)
    return OnnxVar(cs, diff, op=OnnxAdd)


def argmax(x, axis=0, keepdims=0):
    """
    See :epkg:`numpy:argmax`.

    .. warning::
        ONNX does not implement default value axis=None.
    """
    if axis is None:
        raise NotImplementedError(  # pragma: no cover
            "ONNX does not allow axis=None.")
    return OnnxVar(x, op=OnnxArgMax, axis=axis, keepdims=keepdims)


def argmin(x, axis=0, keepdims=0):
    """
    See :epkg:`numpy:argmin`.

    .. warning::
        ONNX does not implement default value axis=None.
    """
    if axis is None:
        raise NotImplementedError(  # pragma: no cover
            "ONNX does not allow axis=None.")
    return OnnxVar(x, op=OnnxArgMin, axis=axis, keepdims=keepdims)


def asin(x):
    "See :epkg:`numpy:asin`."
    return OnnxVar(x, op=OnnxAsin)


def asinh(x):
    "See :epkg:`numpy:asinh`."
    return OnnxVar(x, op=OnnxAsinh)


def atan(x):
    "See :epkg:`numpy:atan`."
    return OnnxVar(x, op=OnnxAtan)


def atanh(x):
    "See :epkg:`numpy:atanh`."
    return OnnxVar(x, op=OnnxAtanh)


def ceil(x):
    "See :epkg:`numpy:ceil`."
    return OnnxVar(x, op=OnnxCeil)


def clip(x, a_min=None, a_max=None):
    "See :epkg:`numpy:clip`."
    args = [x]
    if a_min is not None:
        args.append(a_min)
    if a_max is not None:
        args.append(a_max)
    return OnnxVar(*args, op=OnnxClip)


def compress(condition, x, axis=None):
    "See :epkg:`numpy:compress`."
    if axis is None:
        return OnnxVar(x, condition, op=OnnxCompress)
    return OnnxVar(x, condition, op=OnnxCompress, axis=axis)


def cos(x):
    "See :epkg:`numpy:cos`."
    return OnnxVar(x, op=OnnxCos)


def cosh(x):
    "See :epkg:`numpy:cosh`."
    return OnnxVar(x, op=OnnxCosh)


def concat(*x, axis=0):
    """
    Operator concat, handle :epkg:`numpy:vstack` and
    :epkg:`numpy:hstack`.
    """
    return OnnxVar(*x, op=OnnxConcat, axis=axis)


def cumsum(x, axis):
    "See :epkg:`numpy:cumsum`."
    return OnnxVar(x, axis, op=OnnxCumSum)


def cst(x):
    """
    Creates a constant. `log(x) + numpy.float32(1)` works
    but `numpy.float32(32) + log(x)` fails because Python
    calls `numpy.float32.__add__` instead of
    `OnnxVar.__add__`. With this function, expression
    `cst(1.) + log(x)` is valid.
    """
    if isinstance(x, float):
        return OnnxVar(numpy.array([x], dtype=numpy.float32),
                       op=OnnxIdentity)
    if isinstance(x, int):
        return OnnxVar(numpy.array([x], dtype=numpy.int64),
                       op=OnnxIdentity)
    if isinstance(x, numpy.ndarray):
        return OnnxVar(x, op=OnnxIdentity)
    if hasattr(x, 'dtype'):
        return OnnxVar(numpy.array([x], dtype=x.dtype),
                       op=OnnxIdentity)
    raise NotImplementedError(
        "Unable to convert type %r into a constant." % type(x))


def det(x):
    "See :epkg:`numpy:linalg:det`."
    return OnnxVar(x, op=OnnxDet)


def dot(a, b):
    "See :epkg:`numpy:dot`."
    return OnnxVar(a, b, op=OnnxMatMul)


def einsum(*x, equation=None):
    "See :epkg:`numpy:einsum`."
    return OnnxVar(*x, op=OnnxEinsum, equation=equation)


def erf(x):
    "See :epkg:`scipy:special:erf`."
    return OnnxVar(x, op=OnnxErf)


def exp(x):
    "See :epkg:`numpy:exp`."
    return OnnxVar(x, op=OnnxExp)


def expand_dims(x, axis):
    "See :epkg:`numpy:expand_dims`."
    if not isinstance(axis, int):
        raise NotImplementedError(  # pragma: no cover
            "This function only allows integer for axis not %r." % type(axis))
    return OnnxVar(x, numpy.array([axis], dtype=numpy.int64),
                   op=OnnxUnsqueeze)


def expit(x):
    "See :epkg:`scipy:special:expit`."
    return OnnxVar(x, op=OnnxSigmoid)


def floor(x):
    "See :epkg:`numpy:floor`."
    return OnnxVar(x, op=OnnxFloor)


def hstack(*x):
    "See :epkg:`numpy:hstack`."
    return OnnxVar(*x, op=OnnxConcat, axis=-1)


def isnan(x):
    "See :epkg:`numpy:isnan`."
    return OnnxVar(x, op=OnnxIsNaN)


def identity(x):
    "Identity."
    return OnnxVar(x, op=OnnxIdentity)


def log(x):
    "See :epkg:`numpy:log`."
    return OnnxVar(x, op=OnnxLog)


def mean(x, axis=None, keepdims=0):
    "See :epkg:`numpy:mean`."
    if axis is None:
        return OnnxVar(x, op=OnnxReduceMean, keepdims=keepdims)
    if not isinstance(axis, list):
        axis = [axis]
    return OnnxVar(x, op=OnnxReduceMean, keepdims=keepdims, axes=axis)


def pad(x, pads, constant_value=None, mode='constant'):
    """
    It does not implement :epkg:`numpy:pad` but the ONNX version
    :func:`onnx_pad <mlprodict.onnxrt.ops_cpu.op_pad.onnx_pad>`.
    """
    if constant_value is None:
        return OnnxVar(x, pads, op=OnnxPad, mode=mode)
    return OnnxVar(x, pads, constant_value, op=OnnxPad, mode=mode)


def prod(x, axis=None, keepdims=0):
    "See :epkg:`numpy:prod`."
    if axis is None:
        return OnnxVar(x, op=OnnxReduceProd, keepdims=keepdims)
    if not isinstance(axis, list):
        axis = [axis]
    return OnnxVar(x, op=OnnxReduceProd, keepdims=keepdims, axes=axis)


def relu(x):
    "relu"
    return OnnxVar(x, op=OnnxRelu)


def reciprocal(x):
    "See :epkg:`numpy:reciprocal`."
    return OnnxVar(x, op=OnnxReciprocal)


def round(x):
    "See :epkg:`numpy:round`."
    return OnnxVar(x, op=OnnxRound)


def sigmoid(x):
    "See :epkg:`scipy:special:expit`."
    return OnnxVar(x, op=OnnxSigmoid)


def sign(x):
    "See :epkg:`numpy:sign`."
    return OnnxVar(x, op=OnnxSign)


def sin(x):
    "See :epkg:`numpy:sin`."
    return OnnxVar(x, op=OnnxSin)


def sinh(x):
    "See :epkg:`numpy:sinh`."
    return OnnxVar(x, op=OnnxSinh)


def sqrt(x):
    "See :epkg:`numpy:sqrt`."
    return OnnxVar(x, op=OnnxSqrt)


def squeeze(x, axis=None):
    "See :epkg:`numpy:squeeze`."
    if axis is None:
        raise NotImplementedError(  # pragma: no cover
            "The case where all empty dimensions are removed is not "
            "implemented.")
    if isinstance(axis, int):
        raise RuntimeError(  # pragma: no cover
            "axis must be a tensor.")
    return OnnxVar(x, axis, op=OnnxSqueeze)


def sum(x, axis=None, keepdims=0):
    "See :epkg:`numpy:sum`."
    if axis is None:
        return OnnxVar(x, op=OnnxReduceSum, keepdims=keepdims)
    return OnnxVar(x, numpy.array([axis], dtype=numpy.int64),
                   op=OnnxReduceSum, keepdims=keepdims)


def tan(x):
    "See :epkg:`numpy:tan`."
    return OnnxVar(x, op=OnnxTan)


def tanh(x):
    "See :epkg:`numpy:tanh`."
    return OnnxVar(x, op=OnnxTanh)


def topk(x, k, axis=-1, largest=1, sorted=1):
    "See :epkg:`numpy:argsort`."
    return xtuple(x, k, op=OnnxTopK, axis=axis, largest=largest,
                  sorted=sorted)


def transpose(x, perm=(1, 0)):
    "See :epkg:`numpy:transpose`."
    return OnnxVar(x, op=OnnxTranspose, perm=list(perm))


def unsqueeze(x, axes):
    "See :epkg:`numpy:expand_dims`."
    return OnnxVar(x, axes, op=OnnxUnsqueeze)


def vstack(*x):
    "See :epkg:`numpy:vstack`."
    return OnnxVar(*x, op=OnnxConcat, axis=0)


def where(cond, x, y):
    "See :epkg:`numpy:where`."
    return OnnxVar(cond, x, y, op=OnnxWhere)



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *ops_cpu*.
"""
from ._op import OpRunOnnxEmpty


def load_op(onnx_node, desc=None, options=None, variables=None, dtype=None):
    """
    Gets the operator related to the *onnx* node.
    This runtime does nothing and never complains.

    :param onnx_node: :epkg:`onnx` node
    :param desc: internal representation
    :param options: runtime options
    :param variables: registered variables created by previous operators
    :param dtype: float computation type
    :return: runtime class
    """
    if desc is None:
        raise ValueError(  # pragma: no cover
            "desc should not be None.")
    return OpRunOnnxEmpty(onnx_node, desc, variables=variables,
                          dtype=dtype, **options)



"""
@file
@brief Parsers for LightGBM booster.
"""
import numpy
from sklearn.base import ClassifierMixin
from skl2onnx._parse import _parse_sklearn_classifier, _parse_sklearn_simple_model
from skl2onnx.common._apply_operation import apply_concat, apply_cast
from skl2onnx.common.data_types import guess_proto_type
from skl2onnx.proto import onnx_proto


class WrappedLightGbmBooster:
    """
    A booster can be a classifier, a regressor.
    Trick to wrap it in a minimal function.
    """

    def __init__(self, booster):
        self.booster_ = booster
        self._model_dict = self.booster_.dump_model()
        self.classes_ = self._generate_classes(self._model_dict)
        self.n_features_ = len(self._model_dict['feature_names'])
        if self._model_dict['objective'].startswith('binary'):
            self.operator_name = 'LgbmClassifier'
        elif self._model_dict['objective'].startswith('regression'):  # pragma: no cover
            self.operator_name = 'LgbmRegressor'
        else:  # pragma: no cover
            raise NotImplementedError('Unsupported LightGbm objective: {}'.format(
                self._model_dict['objective']))
        if self._model_dict.get('average_output', False):
            self.boosting_type = 'rf'
        else:
            # Other than random forest, other boosting types do not affect later conversion.
            # Here `gbdt` is chosen for no reason.
            self.boosting_type = 'gbdt'

    @staticmethod
    def _generate_classes(model_dict):
        if model_dict['num_class'] == 1:
            return numpy.asarray([0, 1])
        return numpy.arange(model_dict['num_class'])


class WrappedLightGbmBoosterClassifier(ClassifierMixin):
    """
    Trick to wrap a LGBMClassifier into a class.
    """

    def __init__(self, wrapped):  # pylint: disable=W0231
        for k in {'boosting_type', '_model_dict', 'operator_name',
                  'classes_', 'booster_', 'n_features_'}:
            setattr(self, k, getattr(wrapped, k))


class MockWrappedLightGbmBoosterClassifier(WrappedLightGbmBoosterClassifier):
    """
    Mocked lightgbm.
    """

    def __init__(self, tree):  # pylint: disable=W0231
        self.dumped_ = tree

    def dump_model(self):
        "mock dump_model method"
        self.visited = True
        return self.dumped_


def lightgbm_parser(scope, model, inputs, custom_parsers=None):
    """
    Agnostic parser for LightGBM Booster.
    """
    if hasattr(model, "fit"):
        raise TypeError(  # pragma: no cover
            "This converter does not apply on type '{}'."
            "".format(type(model)))

    if len(inputs) == 1:
        wrapped = WrappedLightGbmBooster(model)
        if wrapped._model_dict['objective'].startswith('binary'):
            wrapped = WrappedLightGbmBoosterClassifier(wrapped)
            return _parse_sklearn_classifier(
                scope, wrapped, inputs, custom_parsers=custom_parsers)
        if wrapped._model_dict['objective'].startswith('regression'):  # pragma: no cover
            return _parse_sklearn_simple_model(
                scope, wrapped, inputs, custom_parsers=custom_parsers)
        raise NotImplementedError(  # pragma: no cover
            "Objective '{}' is not implemented yet.".format(
                wrapped._model_dict['objective']))

    # Multiple columns
    this_operator = scope.declare_local_operator('LightGBMConcat')
    this_operator.raw_operator = model
    this_operator.inputs = inputs
    var = scope.declare_local_variable(
        'Xlgbm', inputs[0].type.__class__([None, None]))
    this_operator.outputs.append(var)
    return lightgbm_parser(scope, model, this_operator.outputs,
                           custom_parsers=custom_parsers)


def shape_calculator_lightgbm_concat(operator):
    """
    Shape calculator for operator *LightGBMConcat*.
    """
    pass


def converter_lightgbm_concat(scope, operator, container):
    """
    Converter for operator *LightGBMConcat*.
    """
    op = operator.raw_operator
    options = container.get_options(op, dict(cast=False))
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:  # pylint: disable=E1101
        proto_dtype = onnx_proto.TensorProto.FLOAT  # pylint: disable=E1101
    if options['cast']:
        concat_name = scope.get_unique_variable_name('cast_lgbm')
        apply_cast(scope, concat_name, operator.outputs[0].full_name, container,
                   operator_name=scope.get_unique_operator_name('cast_lgmb'),
                   to=proto_dtype)
    else:
        concat_name = operator.outputs[0].full_name

    apply_concat(scope, [_.full_name for _ in operator.inputs],
                 concat_name, container, axis=1)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from scipy.linalg import solve
from ._op import OpRunBinaryNum
from ._new_ops import OperatorSchema


class Solve(OpRunBinaryNum):

    atts = {'lower': False,
            'transposed': False}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunBinaryNum.__init__(self, onnx_node, desc=desc,
                                expected_attributes=Solve.atts,
                                **options)

    def _find_custom_operator_schema(self, op_name):
        if op_name == "Solve":
            return SolveSchema()
        raise RuntimeError(  # pragma: no cover
            "Unable to find a schema for operator '{}'.".format(op_name))

    def _run(self, a, b):  # pylint: disable=W0221
        if self.inplaces.get(1, False):
            return (solve(a, b, overwrite_b=True, lower=self.lower,
                          transposed=self.transposed), )
        return (solve(a, b, lower=self.lower, transposed=self.transposed), )

    def _infer_shapes(self, a, b):  # pylint: disable=W0221,W0237
        return (b, )

    def _infer_types(self, a, b):  # pylint: disable=W0221,W0237
        return (b, )

    def to_python(self, inputs):
        return ('from scipy.linalg import solve',
                "return solve({}, {}, lower={}, transposed={})".format(
                    inputs[0], inputs[1], self.lower, self.transposed))


class SolveSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl TreeEnsembleClassifierDouble.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'Solve')
        self.attributes = Solve.atts



"""
@file
@brief Wrapper around :epkg:`onnxruntime`.

.. versionadded:: 0.6
"""
import os
from onnx import numpy_helper

try:
    from onnxruntime import (  # pylint: disable=W0611
        SessionOptions, RunOptions,
        InferenceSession as OrtInferenceSession,
        __version__ as onnxrt_version,
        GraphOptimizationLevel)
except ImportError:
    SessionOptions = None
    RunOptions = None
    OrtInferenceSession = None
    onnxrt_version = "0.0.0"
    GraphOptimizationLevel = None

try:
    from onnxruntime.capi.onnxruntime_pybind11_state import (  # pylint: disable=W0611
        Fail as OrtFail,
        NotImplemented as OrtNotImplemented,
        InvalidArgument as OrtInvalidArgument,
        InvalidGraph as OrtInvalidGraph,
        RuntimeException as OrtRuntimeException)
except ImportError:
    SessionOptions = None
    RunOptions = None
    InferenceSession = None
    onnxrt_version = "0.0.0"
    GraphOptimizationLevel = None
    OrtFail = RuntimeError
    OrtNotImplemented = RuntimeError
    OrtInvalidArgument = RuntimeError
    OrtInvalidGraph = RuntimeError
    OrtRuntimeException = RuntimeError


class InferenceSession:  # pylint: disable=E0102
    """
    Wrappers around InferenceSession from :epkg:`onnxruntime`.

    :param onnx_bytes: onnx bytes
    :param session_options: session options
    """

    def __init__(self, onnx_bytes, sess_options=None, log_severity_level=4):
        if InferenceSession is None:
            raise ImportError(  # pragma: no cover
                "onnxruntime is not available.")
        self.log_severity_level = log_severity_level
        if sess_options is None:
            self.so = SessionOptions()
            self.so.log_severity_level = log_severity_level
            self.sess = OrtInferenceSession(onnx_bytes, sess_options=self.so)
        else:
            self.sess = OrtInferenceSession(
                onnx_bytes, sess_options=sess_options)
        self.ro = RunOptions()
        self.ro.log_severity_level = log_severity_level

    def run(self, output_names, input_feed, run_options=None):
        """
        Executes the ONNX graph.

        :param output_names: None for all, a name for a specific output
        :param input_feed: dictionary of inputs
        :param run_options: None or RunOptions
        :return: array
        """
        return self.sess.run(output_names, input_feed, run_options or self.ro)

    def get_inputs(self):
        "Returns input types."
        return self.sess.get_inputs()

    def get_outputs(self):
        "Returns output types."
        return self.sess.get_outputs()

    def end_profiling(self):
        "Ends profiling."
        return self.sess.end_profiling()


def prepare_c_profiling(model_onnx, inputs, dest=None):
    """
    Prepares model and data to be profiled with tool `perftest
    <https://github.com/microsoft/onnxruntime/tree/
    master/onnxruntime/test/perftest>`_ (onnxruntime) or
    `onnx_test_runner <https://github.com/microsoft/
    onnxruntime/blob/master/docs/Model_Test.md>`_.
    It saves the model in folder
    *dest* and dumps the inputs in a subfolder.

    :param model_onnx: onnx model
    :param inputs: inputs as a list of a dictionary
    :param dest: destination folder, None means the current folder
    :return: command line to use
    """
    if dest is None:
        dest = "."
    if not os.path.exists(dest):
        os.makedirs(dest)
    dest = os.path.abspath(dest)
    name = "model.onnx"
    model_bytes = model_onnx.SerializeToString()
    with open(os.path.join(dest, name), "wb") as f:
        f.write(model_bytes)
    sess = InferenceSession(model_bytes)
    input_names = [_.name for _ in sess.get_inputs()]
    if isinstance(inputs, list):
        dict_inputs = dict(zip(input_names, inputs))
    else:
        dict_inputs = inputs
        inputs = [dict_inputs[n] for n in input_names]
    outputs = sess.run(None, dict_inputs)
    sub = os.path.join(dest, "test_data_set_0")
    if not os.path.exists(sub):
        os.makedirs(sub)
    for i, v in enumerate(inputs):
        n = os.path.join(sub, "input_%d.pb" % i)
        pr = numpy_helper.from_array(v)
        with open(n, "wb") as f:
            f.write(pr.SerializeToString())
    for i, v in enumerate(outputs):
        n = os.path.join(sub, "output_%d.pb" % i)
        pr = numpy_helper.from_array(v)
        with open(n, "wb") as f:
            f.write(pr.SerializeToString())

    cmd = 'onnx_test_runner -e cpu -r 100 -c 1 "%s"' % dest
    return cmd



"""
@file
@brief Complex but recurring testing functions.
"""
import random
import pandas
import numpy
from numpy.testing import assert_allclose
from ..grammar_sklearn import sklearn2graph
from ..grammar_sklearn.cc import compile_c_function


def iris_data():
    """
    Returns ``(X, y)`` for iris data.
    """
    from sklearn.datasets import load_iris
    iris = load_iris()
    X = iris.data[:, :2]
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    y = iris.target
    return X, y


def check_is_almost_equal(xv, exp, precision=1e-5, message=None):
    """
    Checks that two floats or two arrays are almost equal.

    @param  xv          float or vector
    @param  exp         expected value
    @param  precision   precision
    @param  message     additional message
    """
    if isinstance(exp, float) or len(exp.ravel()) == 1:
        if not (isinstance(xv, float) or len(xv.ravel()) == 1):
            raise TypeError(  # pragma: no cover
                "Type mismatch between {0} and {1} (expected).".format(
                    type(xv), type(exp)))
        diff = abs(xv - exp)
        if diff > 1e-5:
            raise ValueError(  # pragma: no cover
                "Predictions are different expected={0}, computed={1}".format(
                    exp, xv))
    else:
        if not isinstance(xv, numpy.ndarray):
            raise TypeError(
                "Type mismatch between {0} and {1} (expected).".format(type(xv), type(exp)))
        xv = xv.ravel()
        exp = exp.ravel()
        try:
            assert_allclose(xv, exp, atol=precision)
        except AssertionError as e:
            if message is None:
                raise e
            else:
                raise AssertionError(message) from e  # pragma: no cover


def check_model_representation(model, X, y=None, convs=None,
                               output_names=None, only_float=True,
                               verbose=False, suffix="", fLOG=None):
    """
    Checks that a trained model can be exported in a specific list
    of formats and produces the same outputs if the
    representation can be used to predict.

    @param  model           model (a class or an instance of a model but not trained)
    @param  X               features
    @param  y               targets
    @param  convs           list of format to check, all possible by default ``['json', 'c']``
    @param  output_names    list of output columns
                            (can be None, a default value is infered based on scikit-learn output then)
    @param  verbose         print some information
    @param  suffix          add this to disambiguate module
    @param  fLOG            logging function
    @return                 function to call to run the prediction
    """
    if not only_float:
        raise NotImplementedError(  # pragma: no cover
            "Only float are allowed.")
    if isinstance(X, list):
        X = pandas.DataFrame(X)
        if len(X.shape) != 2:
            raise ValueError(  # pragma: no cover
                "X cannot be converted into a proper DataFrame. It has shape {0}."
                "".format(X.shape))
        if only_float:
            X = X.values
    if isinstance(y, list):
        y = numpy.array(y)
    if convs is None:
        convs = ['json', 'c']

    # sklearn
    if not hasattr(model.__class__, "fit"):
        # It is a class object and not an instance.
        # We use the default values.
        model = model()

    model.fit(X, y)
    h = random.randint(0, X.shape[0] - 1)
    if isinstance(X, pandas.DataFrame):
        oneX = X.iloc[h, :].astype(numpy.float32)
    else:
        oneX = X[h, :].ravel().astype(numpy.float32)

    # model or transform
    moneX = numpy.resize(oneX, (1, len(oneX)))
    if hasattr(model, "predict"):
        ske = model.predict(moneX)
    else:
        ske = model.transform(moneX)

    if verbose and fLOG:
        fLOG("---------------------")
        fLOG(type(oneX), oneX.dtype)
        fLOG(model)
        for k, v in sorted(model.__dict__.items()):
            if k[-1] == '_':
                fLOG("  {0}={1}".format(k, v))
        fLOG("---------------------")

    # grammar
    gr = sklearn2graph(model, output_names=output_names)
    lot = gr.execute(Features=oneX)
    if verbose and fLOG:
        fLOG(gr.graph_execution())

    # verification
    check_is_almost_equal(lot, ske)

    # default for output_names
    if output_names is None:
        if len(ske.shape) == 1:
            output_names = ["Prediction"]
        elif len(ske.shape) == 2:
            output_names = ["p%d" % i for i in range(ske.shape[1])]
        else:
            raise ValueError(  # pragma: no cover
                "Cannot guess default values for output_names.")

    for lang in convs:
        if lang in ('c', ):
            code_c = gr.export(lang=lang)['code']
            if code_c is None:
                raise ValueError("cannot be None")  # pragma: no cover

            compile_fct = compile_c_function

            from contextlib import redirect_stdout, redirect_stderr
            from io import StringIO
            fout = StringIO()
            ferr = StringIO()
            with redirect_stdout(fout):
                with redirect_stderr(ferr):
                    try:
                        fct = compile_fct(
                            code_c, len(output_names), suffix=suffix, fLOG=lambda s: fout.write(s + "\n"))
                    except Exception as e:  # pragma: no cover
                        raise RuntimeError(
                            "Unable to compile a code\n-OUT-\n{0}\n-ERR-\n{1}\n-CODE-"
                            "\n{2}".format(fout.getvalue(), ferr.getvalue(), code_c)) from e

            if verbose and fLOG:
                fLOG("-----------------")
                fLOG(output_names)
                fLOG("-----------------")
                fLOG(code_c)
                fLOG("-----------------")
                fLOG("h=", h, "oneX=", oneX)
                fLOG("-----------------")
            lotc = fct(oneX)
            check_is_almost_equal(
                lotc, ske, message="Issue with lang='{0}'".format(lang))
            lotc_exp = lotc.copy()
            lotc2 = fct(oneX, lotc)
            if not numpy.array_equal(lotc_exp, lotc2):
                raise ValueError(  # pragma: no cover
                    "Second call returns different results.\n{0}\n{1}".format(
                        lotc_exp, lotc2))
        else:
            ser = gr.export(lang="json", hook={'array': lambda v: v.tolist()})
            if ser is None:
                raise ValueError(  # pragma: no cover
                    "No output for long='{0}'".format(lang))



"""
@file
@brief Functions to help get more information about the models.
"""
import inspect
from collections import Counter
import numpy


def _analyse_tree(tree):
    """
    Extract information from a tree.
    """
    info = {}
    if hasattr(tree, 'node_count'):
        info['node_count'] = tree.node_count

    n_nodes = tree.node_count
    children_left = tree.children_left
    children_right = tree.children_right
    node_depth = numpy.zeros(shape=n_nodes, dtype=numpy.int64)
    is_leaves = numpy.zeros(shape=n_nodes, dtype=bool)
    stack = [(0, -1)]
    while len(stack) > 0:
        node_id, parent_depth = stack.pop()
        node_depth[node_id] = parent_depth + 1
        if children_left[node_id] != children_right[node_id]:
            stack.append((children_left[node_id], parent_depth + 1))
            stack.append((children_right[node_id], parent_depth + 1))
        else:
            is_leaves[node_id] = True

    info['leave_count'] = sum(is_leaves)
    info['max_depth'] = max(node_depth)
    return info


def _analyse_tree_h(tree):
    """
    Extract information from a tree in a
    HistGradientBoosting.
    """
    info = {}
    info['leave_count'] = tree.get_n_leaf_nodes()
    info['node_count'] = len(tree.nodes)
    info['max_depth'] = tree.get_max_depth()
    return info


def _reduce_infos(infos):
    """
    Produces agregates features.
    """
    def tof(obj):
        try:
            return obj[0]
        except TypeError:  # pragma: no cover
            return obj

    if not isinstance(infos, list):
        raise TypeError(  # pragma: no cover
            "infos must a list not {}.".format(type(infos)))
    keys = set()
    for info in infos:
        if not isinstance(info, dict):
            raise TypeError(  # pragma: no cover
                "info must a dictionary not {}.".format(type(info)))
        keys |= set(info)

    info = {}
    for k in keys:
        values = [d.get(k, None) for d in infos]
        values = [_ for _ in values if _ is not None]
        if k.endswith('.leave_count') or k.endswith('.node_count'):
            info['sum|%s' % k] = sum(values)
        elif k.endswith('.max_depth'):
            info['max|%s' % k] = max(values)
        elif k.endswith('.size'):
            info['sum|%s' % k] = sum(values)  # pragma: no cover
        else:
            try:
                un = set(values)
            except TypeError:  # pragma: no cover
                un = set()
            if len(un) == 1:
                info[k] = list(un)[0]
                continue
            if k.endswith('.shape'):
                row = [_[0] for _ in values]
                col = [_[1] for _ in values if len(_) > 1]
                if len(col) == 0:
                    info['max|%s' % k] = (max(row), )
                else:
                    info['max|%s' % k] = (max(row), max(col))
                continue
            if k == 'n_classes_':
                info['n_classes_'] = max(tof(_) for _ in values)
                continue
            raise NotImplementedError(  # pragma: no cover
                "Unable to reduce key '{}', values={}.".format(k, values))
    return info


def _get_info_lgb(model):
    """
    Get informations from and :epkg:`lightgbm` trees.
    """
    from ..onnx_conv.operator_converters.conv_lightgbm import (
        _parse_tree_structure,
        get_default_tree_classifier_attribute_pairs
    )
    gbm_text = model.dump_model()

    info = {'objective': gbm_text['objective']}
    if gbm_text['objective'].startswith('binary'):
        info['n_classes'] = 1
    elif gbm_text['objective'].startswith('multiclass'):
        info['n_classes'] = gbm_text['num_class']
    elif gbm_text['objective'].startswith('regression'):
        info['n_targets'] = 1
    else:
        raise NotImplementedError(  # pragma: no cover
            "Unknown objective '{}'.".format(gbm_text['objective']))
    n_classes = info.get('n_classes', info.get('n_targets', -1))

    info['estimators_.size'] = len(gbm_text['tree_info'])
    attrs = get_default_tree_classifier_attribute_pairs()
    for i, tree in enumerate(gbm_text['tree_info']):
        tree_id = i
        class_id = tree_id % n_classes
        learning_rate = 1.
        _parse_tree_structure(
            tree_id, class_id, learning_rate, tree['tree_structure'], attrs)

    info['node_count'] = len(attrs['nodes_nodeids'])
    info['ntrees'] = len(set(attrs['nodes_treeids']))
    dist = Counter(attrs['nodes_modes'])
    info['leave_count'] = dist['LEAF']
    info['mode_count'] = len(dist)
    return info


def _get_info_xgb(model):
    """
    Get informations from and :epkg:`lightgbm` trees.
    """
    from ..onnx_conv.operator_converters.conv_xgboost import (
        XGBConverter, XGBClassifierConverter)
    objective, _, js_trees = XGBConverter.common_members(model, None)
    attrs = XGBClassifierConverter._get_default_tree_attribute_pairs()
    XGBConverter.fill_tree_attributes(
        js_trees, attrs, [1 for _ in js_trees], True)
    info = {'objective': objective}
    info['estimators_.size'] = len(js_trees)
    info['node_count'] = len(attrs['nodes_nodeids'])
    info['ntrees'] = len(set(attrs['nodes_treeids']))
    dist = Counter(attrs['nodes_modes'])
    info['leave_count'] = dist['LEAF']
    info['mode_count'] = len(dist)
    return info


def analyze_model(model, simplify=True):
    """
    Returns informations, statistics about a model,
    its number of nodes, its size...

    @param      model       any model
    @param      simplify    simplifies the tuple of length 1
    @return                 dictionary

    .. exref::
        :title: Extract information from a model

        The function @see fn analyze_model extracts global
        figures about a model, whatever it is.

        .. runpython::
            :showcode:
            :warningout: DeprecationWarning

            import pprint
            from sklearn.datasets import load_iris
            from sklearn.ensemble import RandomForestClassifier
            from mlprodict.tools.model_info import analyze_model

            data = load_iris()
            X, y = data.data, data.target
            model = RandomForestClassifier().fit(X, y)
            infos = analyze_model(model)
            pprint.pprint(infos)
    """
    if hasattr(model, 'SerializeToString'):
        # ONNX model
        from ..onnx_tools.optim.onnx_helper import onnx_statistics
        return onnx_statistics(model)

    if isinstance(model, numpy.ndarray):
        info = {'shape': model.shape}
        infos = []
        for v in model.ravel():
            if hasattr(v, 'fit'):
                ii = analyze_model(v, False)
                infos.append(ii)
        if len(infos) == 0:
            return info  # pragma: no cover
        for k, v in _reduce_infos(infos).items():
            info['.%s' % k] = v
        return info

    # linear model
    info = {}
    for k in model.__dict__:
        if k in ['tree_']:
            continue
        if k.endswith('_') and not k.startswith('_'):
            v = getattr(model, k)
            if isinstance(v, numpy.ndarray):
                info['%s.shape' % k] = v.shape
            elif isinstance(v, numpy.float64):
                info['%s.shape' % k] = 1
        elif k in ('_fit_X', ):
            v = getattr(model, k)
            info['%s.shape' % k] = v.shape

    # classification
    for f in ['n_classes_', 'n_outputs', 'n_features_']:
        if hasattr(model, f):
            info[f] = getattr(model, f)

    # tree
    if hasattr(model, 'tree_'):
        for k, v in _analyse_tree(model.tree_).items():
            info['tree_.%s' % k] = v

    # tree
    if hasattr(model, 'get_n_leaf_nodes'):
        for k, v in _analyse_tree_h(model).items():
            info['tree_.%s' % k] = v

    # estimators
    if hasattr(model, 'estimators_'):
        info['estimators_.size'] = len(model.estimators_)
        infos = [analyze_model(est, False) for est in model.estimators_]
        for k, v in _reduce_infos(infos).items():
            info['estimators_.%s' % k] = v

    # predictors
    if hasattr(model, '_predictors'):
        info['_predictors.size'] = len(model._predictors)
        infos = []
        for est in model._predictors:
            ii = [analyze_model(e, False) for e in est]
            infos.extend(ii)
        for k, v in _reduce_infos(infos).items():
            info['_predictors.%s' % k] = v

    # LGBM
    if hasattr(model, 'booster_'):
        info.update(_get_info_lgb(model.booster_))

    # XGB
    if hasattr(model, 'get_booster'):
        info.update(_get_info_xgb(model))

    # end
    if simplify:
        up = {}
        for k, v in info.items():
            if isinstance(v, tuple) and len(v) == 1:
                up[k] = v[0]
        info.update(up)

    return info


def enumerate_models(model):
    """
    Enumerates models with models.

    @param      model       :epkg:`scikit-learn` model
    @return                 enumerate models
    """
    yield model
    sig = inspect.signature(model.__init__)
    for k in sig.parameters:
        sub = getattr(model, k, None)
        if sub is None:
            continue
        if not hasattr(sub, 'fit'):
            continue
        for m in enumerate_models(sub):
            yield m


def set_random_state(model, value=0):
    """
    Sets all possible parameter *random_state* to 0.

    @param      model       :epkg:`scikit-learn` model
    @param      value       new value
    @return                 model (same one)
    """
    for m in enumerate_models(model):
        sig = inspect.signature(m.__init__)
        hasit = any(filter(lambda p: p == 'random_state',
                           sig.parameters))
        if hasit and hasattr(m, 'random_state'):
            m.random_state = value
    return model
